/*
 * BRLTTY - A background process providing access to the console screen (when in
 *          text mode) for a blind person using a refreshable braille display.
 *
 * Copyright (C) 1995-2015 by The BRLTTY Developers.
 *
 * BRLTTY comes with ABSOLUTELY NO WARRANTY.
 *
 * This is free software, placed under the terms of the
 * GNU General Public License, as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any
 * later version. Please see the file LICENSE-GPL for details.
 *
 * Web Page: http://brltty.com/
 *
 * This software is maintained by Dave Mielke <dave@mielke.cc>.
 */

#include "prologue.h"

#include <string.h>
#include <errno.h>
#include <stdlib.h>

#ifndef __ANDROID__
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
 #include <X11/keysym.h>
 #include <X11/Xatom.h>
 #include <X11/XKBlib.h>
 #include <X11/extensions/XTest.h>
#endif
#include <pthread.h>
#include "log.h"
#include "ttb.h"
//#include <libusb.h>

#include "io_generic.h"
#include "gio_types.h"
#include "io_usb.h"
#include "usb_internal.h"
#include "gio_internal.h"

#define BRL_HAVE_PACKET_IO

#include "brl_driver.h"
#include "brldefs-pa.h"

#define PROBE_RETRY_LIMIT 2
#define PROBE_INPUT_TIMEOUT 1000
#define MAXIMUM_RESPONSE_SIZE 32

#define MAXIMUM_CELL_COUNT 140

#define PAC_LOG

#ifdef PAC_LOG
#define PAC_LOG_FUNC_BEGIN logMessage(LOG_INFO, "Pactos driver : %s -> begin\n", __func__);

#define PAC_LOG_FUNC_SUCCESS logMessage(LOG_INFO, "Pactos driver : %s -> returned successfully(line %d)\n", __func__, __LINE__);

#define PAC_LOG_FUNC_FAILURE logMessage(LOG_INFO, "Pactos driver : %s -> returned unsuccessfully!(line %d)\n",__func__, __LINE__);
#else
#define PAC_LOG_FUNC_BEGIN
#define PAC_LOG_FUNC_SUCCESS
#define PAC_LOG_FUNC_FAILURE
#endif

//pthread_t table_change_thread ;
FILE *logFile;
FILE *pa_XKeyMap;


// Navigation key definitions
BEGIN_KEY_NAME_TABLE(navigation)
	KEY_NAME_ENTRY(PA_KEY_Up, "Up"),
	KEY_NAME_ENTRY(PA_KEY_Down, "Down"),
	KEY_NAME_ENTRY(PA_KEY_RightScroll, "RightScroll"),
	KEY_NAME_ENTRY(PA_KEY_LeftScroll, "LeftScroll"),
	KEY_GROUP_ENTRY(PA_GRP_RoutingKey, "RoutingKey"),
END_KEY_NAME_TABLE


// Navigation2 key definitions
BEGIN_KEY_NAME_TABLE(navigation2)
	KEY_NAME_ENTRY(PA2_KEY_Up, "Up"),
	KEY_NAME_ENTRY(PA2_KEY_Down, "Down"),
	KEY_NAME_ENTRY(PA2_KEY_RightScroll, "RightScroll"),
	KEY_NAME_ENTRY(PA2_KEY_LeftScroll, "LeftScroll"),
	KEY_NAME_ENTRY(PA2_KEY_Left, "Left"),
	KEY_NAME_ENTRY(PA2_KEY_Right, "Right"),
	KEY_NAME_ENTRY(PA2_KEY_Up2, "Up2"),
	KEY_NAME_ENTRY(PA2_KEY_Down2, "Down2"),
	KEY_GROUP_ENTRY(PA_GRP_RoutingKey, "RoutingKey"),
END_KEY_NAME_TABLE


// Novem key definitions
BEGIN_KEY_NAME_TABLE(novem)
	KEY_NAME_ENTRY(PA_KEY_Dot1, "Dot1"),
	KEY_NAME_ENTRY(PA_KEY_Dot2, "Dot2"),
	KEY_NAME_ENTRY(PA_KEY_Dot3, "Dot3"),
	KEY_NAME_ENTRY(PA_KEY_Dot4, "Dot4"),
	KEY_NAME_ENTRY(PA_KEY_Dot5, "Dot5"),
	KEY_NAME_ENTRY(PA_KEY_Dot6, "Dot6"),
	KEY_NAME_ENTRY(PA_KEY_BackSpace, "BackSpace"),
	KEY_NAME_ENTRY(PA_KEY_Space, "Space"),
END_KEY_NAME_TABLE


// Novem2 key definitions
BEGIN_KEY_NAME_TABLE(novem2)
	KEY_NAME_ENTRY(PA2_KEY_Dot1, "Dot1"),
	KEY_NAME_ENTRY(PA2_KEY_Dot2, "Dot2"),
	KEY_NAME_ENTRY(PA2_KEY_Dot3, "Dot3"),
	KEY_NAME_ENTRY(PA2_KEY_Dot4, "Dot4"),
	KEY_NAME_ENTRY(PA2_KEY_Dot5, "Dot5"),
	KEY_NAME_ENTRY(PA2_KEY_Dot6, "Dot6"),
	KEY_NAME_ENTRY(PA2_KEY_BackSpace, "BackSpace"),
	KEY_NAME_ENTRY(PA2_KEY_Space, "Space"),
	KEY_NAME_ENTRY(PA2_KEY_Enter, "Enter"),
END_KEY_NAME_TABLE


// Control key definitions
BEGIN_KEY_NAME_TABLE(control)
	KEY_NAME_ENTRY(PA2_KEY_ScreenReader,"ScreenReader"),
	KEY_NAME_ENTRY(PA2_KEY_Control, "Control"),
	KEY_NAME_ENTRY(PA2_KEY_Alt, "Alt"),
END_KEY_NAME_TABLE


// BJ3 keys
BEGIN_KEY_NAME_TABLES(BJ_3)
	KEY_NAME_TABLE(navigation),
	KEY_NAME_TABLE(novem),
END_KEY_NAME_TABLES


// NoteTaker keys
BEGIN_KEY_NAME_TABLES(NoteTaker)
	KEY_NAME_TABLE(navigation2),
	KEY_NAME_TABLE(novem2),
	KEY_NAME_TABLE(control),
END_KEY_NAME_TABLES

// Define key tables for different models
DEFINE_KEY_TABLE(NoteTaker)
DEFINE_KEY_TABLE(BJ_3)

BEGIN_KEY_TABLE_LIST
	&KEY_TABLE_DEFINITION(NoteTaker),
	&KEY_TABLE_DEFINITION(BJ_3),
END_KEY_TABLE_LIST



struct GioHandleStruct {
	UsbChannel *channel;
	GioUsbConnectionProperties properties;
};


typedef struct {
	unsigned char identifier;
	unsigned char cellCount;
	const KeyTableDefinition *ktd;
} ModelEntry;

static const ModelEntry modelTable[] = {

{ .identifier = 0x03, .cellCount = 20, .ktd = &KEY_TABLE_DEFINITION(BJ_3) },
{ .identifier = 0x04, .cellCount = 20, .ktd = &KEY_TABLE_DEFINITION(NoteTaker) },

{ .identifier = 0 } };

struct BrailleDataStruct {
	const ModelEntry *model;
	unsigned char forceRewrite;
	unsigned char acknowledgementPending;
	unsigned char textCells[MAXIMUM_CELL_COUNT];
};


static const ModelEntry *
getModelEntry(unsigned char identifier) {
	const ModelEntry *model = modelTable;

	PAC_LOG_FUNC_BEGIN

	while (model->identifier) {
		if (identifier == model->identifier) {
			PAC_LOG_FUNC_SUCCESS

			return model;
		}
		model += 1;
	}
	logMessage(LOG_WARNING, "unknown %s model: 0X%02X", STRINGIFY(DRIVER_NAME),
			identifier);

	PAC_LOG_FUNC_FAILURE
	return NULL;
}

static int setModel(BrailleDisplay *brl, unsigned char identifier) {
	const ModelEntry *model = getModelEntry(identifier);

	PAC_LOG_FUNC_BEGIN

	if (model) {
		logMessage(LOG_NOTICE, "%s Model: 0X%02X, %u cells",
				STRINGIFY(DRIVER_NAME), model->identifier, model->cellCount);

		brl->data->model = model;
		brl->textColumns = model->cellCount;
		PAC_LOG_FUNC_SUCCESS
		return 1;
	}
	PAC_LOG_FUNC_FAILURE
	return 0;
}

static int writeBytes(BrailleDisplay *brl, const unsigned char *bytes,
		size_t count) {
	PAC_LOG_FUNC_BEGIN

	return writeBraillePacket(brl, brl->gioEndpoint, bytes, count);
}

static int writePacket(BrailleDisplay *brl , size_t size, const unsigned char *data) {

	PAC_LOG_FUNC_BEGIN

	unsigned char bytes[size];
	unsigned char *byte = bytes;
	byte = memcpy(byte, data, size);

	return writeBytes(brl, bytes, size);
}

static BraillePacketVerifierResult verifyPacket(BrailleDisplay *brl,
		const unsigned char *bytes, size_t size, size_t *length, void *data) {
	PAC_LOG_FUNC_BEGIN

	*length = PA_USB_BUFFER_SIZE;
	return BRL_PVR_INCLUDE;

}

static size_t readPacket(BrailleDisplay *brl, void *packet, size_t size) {
	PAC_LOG_FUNC_BEGIN

	return readBraillePacket(brl, brl->gioEndpoint, packet, size, verifyPacket,
			NULL);

}

static int connectResource(BrailleDisplay *brl, const char *identifier) {
	PAC_LOG_FUNC_BEGIN

	static const SerialParameters serialParameters = {
			SERIAL_DEFAULT_PARAMETERS,
			.baud = 57600,
			.parity = SERIAL_PARITY_NONE
	};

	BEGIN_USB_CHANNEL_DEFINITIONS
		{ /* NoteTaker */
			.vendor=0x03eb, .product=0x2018,
			.configuration=1, .interface=1, .alternative=0,
			.inputEndpoint=0x03, .outputEndpoint=0x04,
			.serial = &serialParameters
		},
	END_USB_CHANNEL_DEFINITIONS

	GioDescriptor descriptor;
	gioInitializeDescriptor(&descriptor);

	descriptor.usb.channelDefinitions = usbChannelDefinitions;

	descriptor.bluetooth.channelNumber = 1;

	if (connectBrailleResource(brl, identifier, &descriptor, NULL)) {
		PAC_LOG_FUNC_SUCCESS
		return 1;
	}

	PAC_LOG_FUNC_FAILURE
	return 0;
}

static int writeIdentityRequest(BrailleDisplay *brl) {

	PAC_LOG_FUNC_BEGIN

	unsigned char bytes[PA_USB_BUFFER_SIZE] = { 0 };
	bytes[PA_USB_BUFFER_SIZE - 1] = PA_REQ_IDENTITY;
	return writeBytes(brl, bytes, sizeof(bytes));

}

static BrailleResponseResult isIdentityResponse(BrailleDisplay *brl,
		const void *packet, size_t size) {

	PAC_LOG_FUNC_BEGIN
	unsigned char *bytes = (unsigned char *) packet;
	int i;
	printf("isIdentity : size ---> %d\n", size);
	for (i = 0; i < 32; i++)
		printf(" %02x", bytes[i]);
	printf("\n");
	if (bytes[size - 1] == 0x82) {
		printf("OOOOOOOOOOOOOOOOOOOOOOOKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKKK\n");
		return BRL_RSP_DONE;
	} else {
		return BRL_RSP_UNEXPECTED;
	}

	PAC_LOG_FUNC_SUCCESS
}

static ssize_t brl_readPacket (BrailleDisplay *brl, void *buffer, size_t size)
{
	return readPacket(brl, buffer, size);
}

static ssize_t brl_writePacket (BrailleDisplay *brl, const void *buffer, size_t size)
{
	return writePacket(brl, size, buffer);
}

static int brl_reset (BrailleDisplay *brl) {}

static int brl_construct(BrailleDisplay *brl, char **parameters,
		const char *device) {

	PAC_LOG_FUNC_BEGIN

	if ((brl->data = malloc(sizeof(*brl->data)))) {
		memset(brl->data, 0, sizeof(*brl->data));

		if (connectResource(brl, device)) {

			unsigned char response[MAXIMUM_RESPONSE_SIZE];

			if (probeBrailleDisplay(brl, PROBE_RETRY_LIMIT, brl->gioEndpoint,
					PROBE_INPUT_TIMEOUT, writeIdentityRequest, readPacket,
					&response, sizeof(response), isIdentityResponse)) {
				if (setModel(brl, 0x04/*response[0]*/)) {
					setBrailleKeyTable(brl, brl->data->model->ktd);
					makeOutputTable(dotsTable_ISO11548_1);

					brl->data->forceRewrite = 1;
					brl->data->acknowledgementPending = 0;

					PAC_LOG_FUNC_SUCCESS
					return 1;
				}
			}
			disconnectBrailleResource(brl, NULL);
		}
		free(brl->data);
	} else {
		logMallocError();
	}

	PAC_LOG_FUNC_FAILURE
	return 0;
}

static void brl_destruct(BrailleDisplay *brl) {

	PAC_LOG_FUNC_BEGIN

	unsigned char reset_msg[32] = { 0 };
	reset_msg[29] = 0x26;
	writeBytes(brl, reset_msg, 32);
	int i, j;
	for (i = 0; i < 1000; i++)
		for (j = 0; j < 1000; j++)
			;

	if (brl->data) {
		free(brl->data);
		brl->data = NULL;
	}
	PAC_LOG_FUNC_SUCCESS
}

static int brl_writeWindow(BrailleDisplay *brl, const wchar_t *text) {

	PAC_LOG_FUNC_BEGIN
	brl->data->acknowledgementPending = 0;
	brl->data->forceRewrite = 0;
	if (!brl->data->acknowledgementPending) {
		if (cellsHaveChanged(brl->data->textCells, brl->buffer,
				brl->textColumns, NULL, NULL, &brl->data->forceRewrite)) {
			unsigned char cells[brl->textColumns];

			translateOutputCells(cells, brl->data->textCells, brl->textColumns);
			if (!writePacket(brl, brl->textColumns, cells)) {
				PAC_LOG_FUNC_FAILURE
				return 0;
			}
			brl->data->acknowledgementPending = 0;

		}
	}

	PAC_LOG_FUNC_SUCCESS
	return 1;
}

int process_packet(BrailleDisplay *brl,	unsigned char *packet)
{
	PAC_LOG_FUNC_BEGIN
	size_t num_bytes = packet[0];

	if ((packet[num_bytes-2] == 0x07) && (packet[num_bytes-1] == 0x00)) //command for changing keyboard layout
	{
		if(strncmp(textTableName, "en_US", strlen("en_US")) == 0)
		{
			logMessage(LOG_INFO,"**************** chnaged to IR *********************");
#ifdef __ANDROID__
			replaceTextTable("/data/data/org.a11y.brltty.android/app_tables", "ir-fa");
			return 1;
#else
			system("setxkbmap  ir");
			replaceTextTable("/etc/brltty", "ir-fa");
#endif
			pa_lang = 1;
			return 1;
		}
		else
		{
			logMessage(LOG_INFO,"**************** chnaged to US *********************");

#ifdef __ANDROID__
			replaceTextTable("/data/data/org.a11y.brltty.android/app_tables", "en_US");
			return 1;
#else
			system("setxkbmap us");
			replaceTextTable("/etc/brltty", "en_US");
#endif
			pa_lang = 0;
			return 1;
		}
	}

	size_t idx = num_bytes-1;
	unsigned char offset = 0;

	logMessage(LOG_DEBUG, "process_packet : before loop");

	// Send keypress events
	while(idx > 0)
	{
		for(int i = 0; i < 8; i++)
		{
			unsigned char ttmmpp =(1 << i) & (packet[idx]) ;
			logMessage(LOG_DEBUG, "ttmmpp == %d", ttmmpp);
			if(ttmmpp )
			{
				enqueueKeyEvent(brl, PA_GRP_NavigationKey, offset+i, 1);
				logMessage(LOG_DEBUG, "process_packet : keypress queued");

			}
		}
		offset += 8;
		idx--;
	}

	idx = num_bytes-1;
	offset = 0;
	// Send keyrelease events
	while(idx > 0)
	{
		for(int i = 0; i < 8; i++)
		{
			unsigned char ttmmpp =(1 << i) & (packet[idx]) ;
			logMessage(LOG_DEBUG, "ttmmpp == %d", ttmmpp);
			if(ttmmpp )
			{
				enqueueKeyEvent(brl, PA_GRP_NavigationKey, offset+i, 0);
				logMessage(LOG_DEBUG, "process_packet : keyrelease queued, offset+i = %d", offset+i);
			}
		}
		offset += 8;
		idx--;
	}

	return 1;
}

static int brl_readCommand(BrailleDisplay *brl, KeyTableCommandContext context) {

	PAC_LOG_FUNC_BEGIN

	unsigned char packet[PA_USB_BUFFER_SIZE];
	size_t size;

	while ((size = readPacket(brl, packet, sizeof(packet))) >= PA_USB_BUFFER_SIZE)
	{
#ifdef PAC_LOG
		logInputPacket(packet, size);
#endif
		if (packet[PA_USB_BUFFER_SIZE - 1] == 0x82 &&
		    packet[0] != 0)
	       	{
			logMessage(LOG_INFO, "Last byte OK!");
			unsigned char key;
			KeyGroup group;

			if (packet[packet[0]] == 0xff)
			{
				process_packet(brl, packet);
			}
			else
			{
				return BRL_CMD_BLK(ROUTE) + packet[packet[0]] - 1;
			}
		}
		logUnexpectedPacket(packet, size);
	}
	return EOF;
}
